"""ExtImage product module."""

#CHANGES FOR CMF COMPATIBILITY: SEARCH FOR "CMF-CHANGE"
#jms/JanixCorp/3.16.2001
#changed tabs to spaces
###############################################################################
#
# Copyright (c) 2000 Gregor Heine and Communications & Design Consultancy GmbH,
# Solmsstrasse 2, 60486 Frankfurt, Germany (http://www.cdc-group.com)
# All rights reserved. Written by Gregor Heine <heine@cdc-group.com>
# ExtFile Home: http://www.zope.org/Members/MacGregor/ExtFile/index_html
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
#  In accordance with the license provided for by the software upon
#  which some of the source code has been derived or used, the following
#  acknowledgement is hereby provided :
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
###############################################################################

__doc__ = """ExtImage product module.
    The ExtImage-Product works like the Zope Image-product, but stores the 
    uploaded image externally in a repository-direcory. It creates a preview
    of the image (requires PIL).

$Id: ExtImage.py,v 2.03 2001/02/26 16:46:35 heine Exp $"""

__version__='$Release: 1.1.0.beta3 $'[10:-2]

import Globals
from Globals import HTMLFile, MessageDialog
from __main__ import *
import Zope
from DateTime import DateTime
from types import IntType, StringType
import urllib, os, string
from os.path import join, isfile
try: from cStringIO import StringIO
except: from StringIO import StringIO
#CMF-CHANGE: had to add full path to 'ExtFile' to avoid grabbing the one in
#'Products' if it had been installed. It happened....
from Products.CMFOptions.ExtFile.ExtFile import *
#from Products.CMFOptions.ExtFile import UNDO_POLICY, NO_BACKUPS, BACKUP_ON_DELETE, ALWAYS_BACKUP, TRANSMAP
NO_PREVIEW = 0
GENERATE = 1
UPLOAD_NORESIZE = 2
UPLOAD_RESIZE = 3

manage_addExtImageForm = HTMLFile('extImageAdd', globals()) 

def manage_addExtImage(self, id='', title='', descr='', file='', preview='', 
                       content_type='', create_prev=0, maxx='', maxy='', 
                       ratio=0, permission_check=0, REQUEST=None):
    """Add a ExtImage to a folder."""
#   print string.split(self._d.absolute_url(),'/')[3:]
    if id=='' and type(file) is not type('') and hasattr(file,'filename'): 
        # generate id from filename and make sure, there are no spaces in the id
        id=file.filename
        id=id[max(string.rfind(id,'/'), 
                  string.rfind(id,'\\'), 
                  string.rfind(id,':')
                 )+1:]
        id=string.translate(id, TRANSMAP)
    if id:
        self=self.this()
        tempExtImage=ExtImage(id, title, descr, permission_check)
        self._setObject(id, tempExtImage)
        if create_prev==UPLOAD_NORESIZE or create_prev==UPLOAD_RESIZE:
            self._getOb(id).manage_file_upload(preview, content_type, 1, create_prev, maxx, maxy, ratio)
            self._getOb(id).manage_file_upload(file, content_type, 0)
        else:
            self._getOb(id).manage_file_upload(file, content_type, 0, create_prev, maxx, maxy, ratio)
        if REQUEST is not None:
            return MessageDialog(title = 'Created',
                message = 'The ExtImage %s was successfully created!' % id,
                action = './manage_main',)
    elif REQUEST is not None:
        return MessageDialog(title = 'Error!',
            message = 'You must specify a file!', action = './manage_main',)

class ExtImage(ExtFile): 
    
    __ac_permissions__=(                # what permissions make sense for us? 
        ('View management screens', ('manage_tabs',
                                     'manage_main',
                                     'manage_uploadForm')),
        ('Change permissions',      ('manage_access',)),
        ('Change ExtFile/ExtImage', ('manage_editExtFile',
                                     'manage_del_prev',
                                     'manage_create_prev',
                                     'manage_file_upload',
                                     'manage_http_upload')),
        ('FTP access',              ('manage_FTPstat',
                                     'manage_FTPget',
                                     'manage_FTPlist')),
        ('Download ExtFile/ExtImage',()),
        ('View',                    ('index_html',
                                     'view',
                                     'icon_html',
                                     'icon_gif',
                                     'preview_html',
                                     'preview',
                                     'link',
                                     'tag',
                                     'is_broken',
                                     'is_webviewable')),
    )

    meta_type = 'ExtImage_CMF'          # what do people think they're adding? 
    
    _image_size={'default':256,'min':10,'max':999} # default,min,max-sizes for the preview image

    ################################
    # Init method                  #
    ################################

    def __init__(self, id, title='', descr='', permission_check=0): 
        """initialize a new instance of ExtImage"""
        ExtImage.inheritedAttribute("__init__")(self, id, title, descr, permission_check)
        # determine the int-values of the size of preview-image
        self.prev_filename=''
        self.prev_content_type=''
        self.has_preview=0

    ################################
    # Public methods               #
    ################################
    
    def __str__(self):
        if self.is_webviewable():
            return self.tag()
        else:
            return self.tag(preview=1)

    def tag(self, preview=0, height=None, width=None, alt=None, 
        scale=0, xscale=0, yscale=0, border="0", **args):
        """
        Generate an HTML IMG tag for this image, with customization.
        Arguments to self.tag() can be any valid attributes of an IMG tag.
        'src' will always be an absolute pathname, to prevent redundant
        downloading of images. Defaults are applied intelligently for
        'height', 'width', and 'alt'. If specified, the 'scale', 'xscale',
        and 'yscale' keyword arguments will be used to automatically adjust
        the output height and width values of the image tag.
        Adopted and adapted from OFS/image.py
        """
        if not self._access_permitted(self.REQUEST): preview=1
        if preview or not self.is_webviewable(): preview='?preview=1'
        else: preview=''
        img_width, img_height = self._getImageSize(self.filename)
        pre_width, pre_height = self._getImageSize(self.prev_filename)
        if height is None: 
            if preview: height=pre_height
            else: height=img_height
        if width is None: 
            if preview: width=pre_width
            else: width=img_width
        
        # Auto-scaling support
        xdelta = xscale or scale
        ydelta = yscale or scale
        if xdelta and width != None:
            width = str(int(width) * xdelta)
        if ydelta and height != None:
            height = str(int(height) * ydelta)
        
        strg='<img src="%s%s" border="%s"' % (self.absolute_url(), preview, border)
        
        if alt is None: alt=self.title_or_id()
        if alt: strg = '%s alt="%s"' % (strg, alt)
        if height: strg = '%s height="%s"' % (strg, height)
        if width: strg = '%s width="%s"' % (strg, width)
        
        for key in args.keys():
            value = args.get(key)
            strg = '%s %s="%s"' % (strg, key, value)
        strg="%s>" % (strg)
        
        if preview and self._access_permitted(self.REQUEST):
            return self.link(strg)
        else:
            return strg
    
    def preview(self, REQUEST, RESPONSE):
        """return a preview of the image"""
        REQUEST.RESPONSE.redirect(REQUEST['URL1']+'?preview=1')
        return ''
    
    def preview_html(self):
        """The preview-image embedded in html with a link to the real file"""
        return self.tag(preview=1)
    
    def is_broken(self):
        """Check if external file exists and return true (1) or false (0)"""
        if self.has_preview and self.filename!=self.prev_filename:
            fn=join(self._reposit_path(),self.prev_filename)
            if not isfile(fn):
                self._undo()
                if not isfile(fn):
                    return 1
        return ExtImage.inheritedAttribute("is_broken")(self)
    
    def is_webviewable(self):
        """return 1 for GIF,JPEG and PNG images, otherwise return 0"""
#CMF-CHANGE:
#changed original line "format=self.format()" due to namespace collide.

        format=self.getFormat()
        
        if format=='JPEG' or format=='GIF' or format=='PNG':
            return 1
        else:
            return 0
    
    def prev_rawsize(self):
        """Returns the size of the preview file."""
        fn=join(self._reposit_path(),self.prev_filename)
        try:
            size=os.stat(fn)[6]
        except:
            try: size=os.stat(fn+'.undo')[6]
            except: size=0
        return size
    
    def prev_size(self):
        """Returns a formatted stringified version of the preview size"""
        return self._bytetostring(self.prev_rawsize())

    def width(self):
        """pixel width of the image"""
        return self._getImageSize(self.filename)[0]
        
    def height(self):
        """pixel height of the image"""
        return self._getImageSize(self.filename)[1]
        
    def prev_width(self):
        """pixel width of the preview"""
        return self._getImageSize(self.prev_filename)[0]
        
    def prev_height(self):
        """pixel height of the preview"""
        return self._getImageSize(self.prev_filename)[1]
        
    def format(self):
        """Get the Fileformat of the image"""
        try:
            from PIL import Image
            im = Image.open(join(self._reposit_path(), self.filename)) 
            return im.format
        except:
            return 'unknown'

    ################################
    # Protected management methods #
    ################################

    #CMF-CHANGE : had to change the dtml file to have a more consistent interface in the native Zope i/f
    manage_main = HTMLFile('extImageEdit', globals())      # Management Interface


    def manage_del_prev(self, REQUEST=None):
        """Delete the Preview Image"""
        if self.has_preview and self.filename!=self.prev_filename:
            prev_fn=join(self._reposit_path(),self.prev_filename)
            try:                                
                if UNDO_POLICY==ALWAYS_BACKUP: os.rename(prev_fn,prev_fn+'.undo')
                else: os.remove(prev_fn)
            except OSError:
                pass
        self.prev_filename=self.prev_content_type=''
        self.has_preview=0
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)

    def manage_create_prev(self, maxx=0, maxy=0, ratio=0, REQUEST=None):
        """Create a preview Image"""
        maxx, maxy=self._formatDimensions(maxx, maxy)
        if maxx!=0 and maxy!=0:
            if self.has_preview and self.filename!=self.prev_filename: self.manage_del_prev()
            try: self._createPreview(self.filename, maxx, maxy, ratio)
            except: self.has_preview=0
            else: self._initPreview()
        if REQUEST is not None:
            if self.has_preview: 
                return self.manage_main(self, REQUEST)
            elif maxx=="0" and maxy=="0":
                return MessageDialog(
                    title = 'Attention',
                    message = "You must enter a value > 0",
                    action = './manage_main',)
            else:
                return MessageDialog(
                    title = 'Warning',
                    message = "An error occurred while generating the preview image.",
                    action = './manage_main',)
                    
    #CMF-CHANGE : had to change the dtml file to have a more consistent interface in the native Zope i/f                
    manage_uploadForm = HTMLFile('extImageUpload', globals())# File upload Interface


    def manage_file_upload(self,  file='', content_type='', is_preview=0, create_prev=0, maxx='', maxy='', ratio=0, REQUEST=None):
        """Upload file from local directory"""
        reposit=self._reposit_path()
        if is_preview and type(file) is not type(''):
            temp_filename=self._get_ufn()+'.preview'
            try:
                fp=open(join(reposit,temp_filename), 'wb')
                self._copy_data(file, fp)
                fp.close()
            except:
                raise 'Cannot write: ', ("%s (%s)" %(self.id, temp_filename))
            if self.has_preview and self.filename!=self.prev_filename: self.manage_del_prev()
            maxx, maxy=self._formatDimensions(maxx, maxy)
            if create_prev==UPLOAD_RESIZE and maxx!=0 and maxy!=0:
                try: self._createPreview(temp_filename, maxx, maxy, ratio)
                except: self.has_preview=0
                else: self._initPreview()
                os.remove(join(reposit, temp_filename))
            else:
                self.prev_filename=temp_filename
                self.prev_content_type=self._get_content_type(file, file.read(100), 
                                  self.id, content_type or self.prev_content_type)
                self.has_preview=1
            self._initPreview()
        else:
            ExtImage.inheritedAttribute("manage_file_upload")(self, file, content_type)
            if create_prev==GENERATE:
                maxx, maxy=self._formatDimensions(maxx, maxy)
                if maxx!=0 and maxy!=0:
                    if self.has_preview and self.filename!=self.prev_filename: self.manage_del_prev()
                    try: self._createPreview(self.filename, maxx, maxy, ratio)
                    except: self.has_preview=0
                    else: self._initPreview()
        if REQUEST is not None:
            return MessageDialog(title = 'Uploaded',
                message = "The file was uploaded successfully!",
                action = './manage_main',)

    def manage_http_upload(self, url, path='', is_preview=0, REQUEST=None):
        """Upload file from http-server."""
        reposit=self._reposit_path()
        if (is_preview):
            url=urllib.quote(url,'/:')
            fp_in=urllib.urlopen(url)
            fp_temp=StringIO()
            self._copy_data(fp_in, fp_temp)
            fp_temp.seek(0)
            prev_filename=self._get_ufn()+'.preview'
            try:
                fp=open(join(reposit,prev_filename), 'wb')
                self._copy_data(fp_temp, fp)
                fp.close()
            except:
                raise 'Cannot write: ', ("%s (%s)" %(self.id, prev_filename))
            if self.has_preview and self.filename!=self.prev_filename: self.manage_del_prev()
            self.prev_filename=prev_filename
            self.prev_content_type=self._get_content_type(fp_temp, fp_temp.read(100), 
                              self.id, self.prev_content_type)
            self.has_preview=1
            self._initPreview()
        else:
            ExtImage.inheritedAttribute("manage_http_upload")(self, url,path)
            if self.has_preview:
                width, height = self._getImageSize(self.prev_filename)
                self.manage_del_prev()
                try: self._createPreview(self.filename, width, height, 0)
                except: self.has_preview=0
                else: self._initPreview()
        if REQUEST is not None:
            return MessageDialog(title = 'Uploaded',
                message = "The file was uploaded successfully!",
                action = './manage_main',)
    
    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests"""
        RESPONSE = ExtImage.inheritedAttribute("PUT")(self,REQUEST, RESPONSE)
        if self.has_preview:
            width, height = self._getImageSize(self.prev_filename)
            self.manage_del_prev()
            try: self._createPreview(self.filename, width, height, 0)
            except: self.has_preview=0
            else: self._initPreview()
        return RESPONSE
    
    ################################
    # Private methods              #
    ################################

    def _initPreview(self):
        prev_width, prev_height=self._getImageSize(self.prev_filename)
        if prev_width<=0 or prev_height<=0: 
            self.has_preview=0
        else:
            self.has_preview=1
            
    def _getImageSize(self, filename):
        #changed this to get the type from the file itself w/o PIL

        #CMF-CHANGE (begin)
        if len(filename) > 0:
            wh = self.getImageInfo(join(self._reposit_path(),filename))
            width = wh[0]
            height = wh[1]
        else:
            width = 0
            height = 0
            
        #CMF-CHANGE (done)      
        
#       try:
#           from PIL import Image
#           im = Image.open(join(self._reposit_path(),filename)) 
#           width=im.size[0]
#           height=im.size[1]
#       except:
#           width=height=0
        return width, height

    def _createPreview(self, filename, maxx, maxy, ratio):
        """generate a preview using PIL"""
        from PIL import Image
        reposit=self._reposit_path()
        im = Image.open(join(reposit, filename)) 
        if im.mode!='RGB' and im.mode!='CMYK': im=im.convert("RGB")
        if ratio==1:            # keep aspect-ratio
            im.thumbnail((maxx,maxy))
        else:                   # distort to fixed size
            im.draft(None,(maxx,maxy))
            im.load()
            im=im.resize((maxx,maxy), Image.BICUBIC)
        prev_filename=self._get_ufn()+'.preview'
        im.save(join(reposit,prev_filename),'JPEG', quality=100)
        self.prev_content_type='image/jpeg'
        self.prev_filename=prev_filename
        self.has_preview=1
    
    def _formatDimensions(self, maxx, maxy):
        """make sure, the dimensions are valid int's"""
        if type(maxx) is StringType:
            try: maxx=string.atoi(maxx)
            except: maxx=self._image_size['default']
        if type(maxy) is StringType:
            try: maxy=string.atoi(maxy)
            except: maxy=self._image_size['default']
        if maxx!=0 and maxy!=0:
            if maxx<self._image_size['min']: maxx=self._image_size['min']
            elif maxx>self._image_size['max']: maxx=self._image_size['max']
            if maxy<self._image_size['min']: maxy=self._image_size['min']
            elif maxy>self._image_size['max']: maxy=self._image_size['max']
        return maxx, maxy
    
    def _undo (self):
        """restore filename after undo or copy-paste"""
        if self.has_preview and self.filename!=self.prev_filename:
            fn=join(self._reposit_path(),self.prev_filename)
            if not isfile(fn) and isfile(fn+'.undo'): 
                os.rename(fn+'.undo', fn)           # rename preview
        return ExtImage.inheritedAttribute("_undo")(self)
    
    ################################
    # Special management methods   #
    ################################

    def manage_afterClone(self, item):
        """When a copy of the object is created (zope copy-paste-operation),
           this function is called by CopySupport.py. A copy of the external file
           is created and self.filename is changed.
        """
        reposit=self._reposit_path()
        new_fn=self._get_ufn()
        if self.has_preview and self.filename!=self.prev_filename:  # rename preview-file
            old_fn=join(reposit,self.prev_filename)
            if (not isfile(old_fn)) and isfile(old_fn+'.undo'): old_fn=old_fn+'.undo'
            if isfile(old_fn):
                try:
                    fp=open(join(reposit,new_fn+'.preview'), 'wb')
                    size=self._copyfile(old_fn, fp)
                    fp.close()
                except:
                    raise "Clone-Exception: ", "Couldn't clone %s.preview" % new_fn
                self.prev_filename=new_fn+'.preview'
            else:
                self.prev_filename=''
                self.has_preview=''
        elif self.has_preview:
            if (isfile(join(reposit,self.filename)) or 
                isfile(join(reposit,self.filename+'.undo'))): 
                self.prev_filename=new_fn
            else:
                self.prev_filename=''
                self.has_preview=''
        else:
            self.prev_filename=''
        return ExtImage.inheritedAttribute ("manage_afterClone") (self, item, new_fn)
        
    def manage_beforeDelete(self, item, container):
        """This method is called, when the object is deleted. To support 
           undo-functionality and because this happens too, when the object 
           is moved (cut-paste) or renamed, the external file is not deleted. 
           It are just renamed to filename.undo and remains in the 
           repository, until it is deleted manually.
        """
        if self.has_preview and self.filename!=self.prev_filename:
            prev_fn=join(self._reposit_path(),self.prev_filename)
            try:                                
                if UNDO_POLICY==NO_BACKUPS: os.remove(prev_fn)
                else: os.rename(prev_fn,prev_fn+'.undo')
            except OSError:
                pass
        return ExtImage.inheritedAttribute ("manage_beforeDelete") (self, item, container)

Globals.default__class_init__(ExtImage)
